from __future__ import annotations

from collections import defaultdict
from typing import Dict, List, Sequence, Tuple, overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client._constants import INSTANCES_LIST_LIMIT_DEFAULT

from cognite.powerops.clients.market_configuration.data_classes import (
    ShopTransformation,
    ShopTransformationApply,
    ShopTransformationList,
)

from ._core import TypeAPI


class ShopTransformationEndsAPI:
    def __init__(self, client: CogniteClient):
        self._client = client

    def retrieve(self, external_id: str | Sequence[str]) -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "power-ops", "externalId": "ShopTransformation.end"},
        )
        if isinstance(external_id, str):
            is_shop_transformation = f.Equals(
                ["edge", "startNode"],
                {"space": "power-ops", "externalId": external_id},
            )
            return self._client.data_modeling.instances.list(
                "edge", limit=-1, filter=f.And(is_edge_type, is_shop_transformation)
            )

        else:
            is_shop_transformations = f.In(
                ["edge", "startNode"],
                [{"space": "power-ops", "externalId": ext_id} for ext_id in external_id],
            )
            return self._client.data_modeling.instances.list(
                "edge", limit=-1, filter=f.And(is_edge_type, is_shop_transformations)
            )

    def list(self, limit=INSTANCES_LIST_LIMIT_DEFAULT) -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "power-ops", "externalId": "ShopTransformation.end"},
        )
        return self._client.data_modeling.instances.list("edge", limit=limit, filter=is_edge_type)


class ShopTransformationStartsAPI:
    def __init__(self, client: CogniteClient):
        self._client = client

    def retrieve(self, external_id: str | Sequence[str]) -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "power-ops", "externalId": "ShopTransformation.start"},
        )
        if isinstance(external_id, str):
            is_shop_transformation = f.Equals(
                ["edge", "startNode"],
                {"space": "power-ops", "externalId": external_id},
            )
            return self._client.data_modeling.instances.list(
                "edge", limit=-1, filter=f.And(is_edge_type, is_shop_transformation)
            )

        else:
            is_shop_transformations = f.In(
                ["edge", "startNode"],
                [{"space": "power-ops", "externalId": ext_id} for ext_id in external_id],
            )
            return self._client.data_modeling.instances.list(
                "edge", limit=-1, filter=f.And(is_edge_type, is_shop_transformations)
            )

    def list(self, limit=INSTANCES_LIST_LIMIT_DEFAULT) -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "power-ops", "externalId": "ShopTransformation.start"},
        )
        return self._client.data_modeling.instances.list("edge", limit=limit, filter=is_edge_type)


class ShopTransformationsAPI(TypeAPI[ShopTransformation, ShopTransformationApply, ShopTransformationList]):
    def __init__(self, client: CogniteClient):
        super().__init__(
            client=client,
            sources=dm.ViewId("power-ops", "ShopTransformation", "9e2166d5480704"),
            class_type=ShopTransformation,
            class_apply_type=ShopTransformationApply,
            class_list=ShopTransformationList,
        )
        self.ends = ShopTransformationEndsAPI(client)
        self.starts = ShopTransformationStartsAPI(client)

    def apply(self, shop_transformation: ShopTransformationApply, replace: bool = False) -> dm.InstancesApplyResult:
        instances = shop_transformation.to_instances_apply()
        return self._client.data_modeling.instances.apply(nodes=instances.nodes, edges=instances.edges, replace=replace)

    def delete(self, external_id: str | Sequence[str]) -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(ShopTransformationApply.space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(ShopTransformationApply.space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> ShopTransformation:
        ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> ShopTransformationList:
        ...

    def retrieve(self, external_id: str | Sequence[str]) -> ShopTransformation | ShopTransformationList:
        if isinstance(external_id, str):
            shop_transformation = self._retrieve((self.sources.space, external_id))

            end_edges = self.ends.retrieve(external_id)
            shop_transformation.end = [edge.end_node.external_id for edge in end_edges]
            start_edges = self.starts.retrieve(external_id)
            shop_transformation.start = [edge.end_node.external_id for edge in start_edges]

            return shop_transformation
        else:
            shop_transformations = self._retrieve([(self.sources.space, ext_id) for ext_id in external_id])

            end_edges = self.ends.retrieve(external_id)
            self._set_end(shop_transformations, end_edges)
            start_edges = self.starts.retrieve(external_id)
            self._set_start(shop_transformations, start_edges)

            return shop_transformations

    def list(self, limit: int = INSTANCES_LIST_LIMIT_DEFAULT) -> ShopTransformationList:
        shop_transformations = self._list(limit=limit)

        end_edges = self.ends.list(limit=-1)
        self._set_end(shop_transformations, end_edges)
        start_edges = self.starts.list(limit=-1)
        self._set_start(shop_transformations, start_edges)

        return shop_transformations

    @staticmethod
    def _set_end(shop_transformations: Sequence[ShopTransformation], end_edges: Sequence[dm.Edge]):
        edges_by_start_node: Dict[Tuple, List] = defaultdict(list)
        for edge in end_edges:
            edges_by_start_node[edge.start_node.as_tuple()].append(edge)

        for shop_transformation in shop_transformations:
            node_id = shop_transformation.id_tuple()
            if node_id in edges_by_start_node:
                shop_transformation.end = [edge.end_node.external_id for edge in edges_by_start_node[node_id]]

    @staticmethod
    def _set_start(shop_transformations: Sequence[ShopTransformation], start_edges: Sequence[dm.Edge]):
        edges_by_start_node: Dict[Tuple, List] = defaultdict(list)
        for edge in start_edges:
            edges_by_start_node[edge.start_node.as_tuple()].append(edge)

        for shop_transformation in shop_transformations:
            node_id = shop_transformation.id_tuple()
            if node_id in edges_by_start_node:
                shop_transformation.start = [edge.end_node.external_id for edge in edges_by_start_node[node_id]]



from __future__ import annotations

from collections import defaultdict
from typing import Dict, List, Sequence, Tuple, overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client._constants import INSTANCES_LIST_LIMIT_DEFAULT

from cognite.powerops.clients.cogshop.data_classes import Mapping, MappingApply, MappingList

from ._core import TypeAPI


class MappingTransformationsAPI:
    def __init__(self, client: CogniteClient):
        self._client = client

    def retrieve(self, external_id: str | Sequence[str]) -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "cogShop", "externalId": "Mapping.transformations"},
        )
        if isinstance(external_id, str):
            is_mapping = f.Equals(
                ["edge", "startNode"],
                {"space": "cogShop", "externalId": external_id},
            )
            return self._client.data_modeling.instances.list("edge", limit=-1, filter=f.And(is_edge_type, is_mapping))

        else:
            is_mappings = f.In(
                ["edge", "startNode"],
                [{"space": "cogShop", "externalId": ext_id} for ext_id in external_id],
            )
            return self._client.data_modeling.instances.list("edge", limit=-1, filter=f.And(is_edge_type, is_mappings))

    def list(self, limit=INSTANCES_LIST_LIMIT_DEFAULT) -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "cogShop", "externalId": "Mapping.transformations"},
        )
        return self._client.data_modeling.instances.list("edge", limit=limit, filter=is_edge_type)


class MappingsAPI(TypeAPI[Mapping, MappingApply, MappingList]):
    def __init__(self, client: CogniteClient):
        super().__init__(
            client=client,
            sources=dm.ViewId("cogShop", "Mapping", "930a8d9497a060"),
            class_type=Mapping,
            class_apply_type=MappingApply,
            class_list=MappingList,
        )
        self.transformations = MappingTransformationsAPI(client)

    def apply(self, mapping: MappingApply, replace: bool = False) -> dm.InstancesApplyResult:
        instances = mapping.to_instances_apply()
        return self._client.data_modeling.instances.apply(nodes=instances.nodes, edges=instances.edges, replace=replace)

    def delete(self, external_id: str | Sequence[str]) -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(MappingApply.space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(MappingApply.space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> Mapping:
        ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> MappingList:
        ...

    def retrieve(self, external_id: str | Sequence[str]) -> Mapping | MappingList:
        if isinstance(external_id, str):
            mapping = self._retrieve((self.sources.space, external_id))

            transformation_edges = self.transformations.retrieve(external_id)
            mapping.transformations = [edge.end_node.external_id for edge in transformation_edges]

            return mapping
        else:
            mappings = self._retrieve([(self.sources.space, ext_id) for ext_id in external_id])

            transformation_edges = self.transformations.retrieve(external_id)
            self._set_transformations(mappings, transformation_edges)

            return mappings

    def list(self, limit: int = INSTANCES_LIST_LIMIT_DEFAULT) -> MappingList:
        mappings = self._list(limit=limit)

        transformation_edges = self.transformations.list(limit=-1)
        self._set_transformations(mappings, transformation_edges)

        return mappings

    @staticmethod
    def _set_transformations(mappings: Sequence[Mapping], transformation_edges: Sequence[dm.Edge]):
        edges_by_start_node: Dict[Tuple, List] = defaultdict(list)
        for edge in transformation_edges:
            edges_by_start_node[edge.start_node.as_tuple()].append(edge)

        for mapping in mappings:
            node_id = mapping.id_tuple()
            if node_id in edges_by_start_node:
                mapping.transformations = [edge.end_node.external_id for edge in edges_by_start_node[node_id]]



import json
from dataclasses import asdict, dataclass
from typing import Dict, Optional

import arrow
import dacite

from cognite.powerops.preprocessor.exceptions import CogShopConfigError
from cognite.powerops.preprocessor.utils import arrow_to_ms, log_and_reraise, rename_dict_keys


@dataclass
class CogShopConfig:
    watercourse: str
    starttime: str
    endtime: str
    timeresolution: Optional[Dict[str, int]] = None
    dynamic_minute_offset: int = 0

    @property
    def starttime_ms(self) -> int:
        return arrow_to_ms(arrow.get(self.starttime))

    @property
    def endtime_ms(self) -> int:
        return arrow_to_ms(arrow.get(self.endtime))

    @classmethod
    @log_and_reraise(CogShopConfigError)
    def from_dict(cls, d: dict, key_prefix="shop:") -> "CogShopConfig":
        with_prefix = {k: v for k, v in d.items() if k.startswith(key_prefix)}
        key_mapping = {key: key.replace(key_prefix, "") for key in with_prefix}
        rename_dict_keys(d=with_prefix, key_mapping=key_mapping)
        if "timeresolution" in with_prefix:  # stringified dict to dict
            with_prefix["timeresolution"] = json.loads(with_prefix["timeresolution"])
        if "dynamic_minute_offset" in with_prefix:  # cast to int
            with_prefix["dynamic_minute_offset"] = int(with_prefix["dynamic_minute_offset"])
        return dacite.from_dict(data_class=cls, data=with_prefix)  # Why not just cls(**with_prefix)

    def to_dict(self, key_prefix="shop:") -> dict:
        without_prefix = asdict(self)
        return {f"{key_prefix}{k}": v for k, v in without_prefix.items() if v is not None}



from __future__ import annotations

import getpass
from pathlib import Path

from cognite.client import ClientConfig, CogniteClient
from cognite.client.credentials import OAuthClientCredentials

from ._api.benchmark_bids import BenchmarkBidsAPI
from ._api.benchmarkings import BenchmarkingsAPI
from ._api.bids import BidsAPI
from ._api.date_transformations import DateTransformationsAPI
from ._api.day_ahead_bids import DayAheadBidsAPI
from ._api.day_ahead_process import DayAheadProcessAPI
from ._api.markets import MarketsAPI
from ._api.nord_pool_markets import NordPoolMarketsAPI
from ._api.production_plan_time_series import ProductionPlanTimeSeriesAPI
from ._api.rkom_bid_combinations import RKOMBidCombinationsAPI
from ._api.rkom_bids import RKOMBidsAPI
from ._api.rkom_markets import RKOMMarketsAPI
from ._api.rkom_process import RKOMProcessAPI
from ._api.shop_transformations import ShopTransformationsAPI


class MarketConfigClient:
    """
    MarketConfigClient

    Generated with:
        pygen = 0.11.4
        cognite-sdk = 6.5.8
        pydantic = 2.0.2

    Data Model:
        space: power-ops
        externalId: MarketConfiguration
        version: 1
    """

    def __init__(self, config: ClientConfig | None = None):
        client = CogniteClient(config)
        self.benchmark_bids = BenchmarkBidsAPI(client)
        self.benchmarkings = BenchmarkingsAPI(client)
        self.bids = BidsAPI(client)
        self.date_transformations = DateTransformationsAPI(client)
        self.day_ahead_bids = DayAheadBidsAPI(client)
        self.day_ahead_process = DayAheadProcessAPI(client)
        self.markets = MarketsAPI(client)
        self.nord_pool_markets = NordPoolMarketsAPI(client)
        self.production_plan_time_series = ProductionPlanTimeSeriesAPI(client)
        self.rkom_bids = RKOMBidsAPI(client)
        self.rkom_bid_combinations = RKOMBidCombinationsAPI(client)
        self.rkom_markets = RKOMMarketsAPI(client)
        self.rkom_process = RKOMProcessAPI(client)
        self.shop_transformations = ShopTransformationsAPI(client)

    @classmethod
    def azure_project(
        cls, tenant_id: str, client_id: str, client_secret: str, cdf_cluster: str, project: str
    ) -> MarketConfigClient:
        base_url = f"https://{cdf_cluster}.cognitedata.com/"
        credentials = OAuthClientCredentials(
            token_url=f"https://login.microsoftonline.com/{tenant_id}/oauth2/v2.0/token",
            client_id=client_id,
            client_secret=client_secret,
            scopes=[f"{base_url}.default"],
        )
        config = ClientConfig(
            project=project,
            credentials=credentials,
            client_name=getpass.getuser(),
            base_url=base_url,
        )

        return cls(config)

    @classmethod
    def from_toml(cls, file_path: Path | str) -> MarketConfigClient:
        import toml

        return cls.azure_project(**toml.load(file_path)["cognite"])



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, ClassVar, Optional

import pandas as pd
from cognite.client.data_classes import FileMetadata, Sequence
from pydantic import BaseModel, ConfigDict

_READ_ONLY_FIELDS = [
    "created_time",
    "last_updated_time",
    "lastUpdatedTime",
    "createdTime",
    "uploaded_time",
    "uploadedTime",
]


class _CDFResource(BaseModel, ABC):
    model_config: ClassVar[ConfigDict] = ConfigDict(arbitrary_types_allowed=True)

    @abstractmethod
    def _dump(self, camel_case: bool = False) -> dict[str, Any]:
        ...

    def dump(self, camel_case: bool = False) -> dict[str, Any]:
        dump = self._dump(camel_case)
        for read_only_field in _READ_ONLY_FIELDS:
            dump.pop(read_only_field, None)
        return dump


class CDFSequence(_CDFResource):
    sequence: Sequence
    content: pd.DataFrame

    @property
    def external_id(self):
        return self.sequence.external_id

    def _set_data_set_id(self, data_set_id: int):
        self.sequence.data_set_id = data_set_id

    data_set_id = property(None, fset=_set_data_set_id)

    def _dump(self, camel_case: bool = False) -> dict[str, Any]:
        return self.sequence.dump(camel_case=camel_case)


class CDFFile(_CDFResource):
    meta: FileMetadata
    content: Optional[bytes] = None

    @property
    def external_id(self):
        return self.meta.external_id

    def _dump(self, camel_case: bool = False) -> dict[str, Any]:
        return self.meta.dump(camel_case=camel_case)
